@@ -53,7 +53,7 @@ static dbus_bool_t qDBusAddTimeout(DBusTimeout *timeout, void *data)
     Q_ASSERT(timeout);
     Q_ASSERT(data);
 
-  //  qDebug("addTimeout %d", dbus_timeout_get_interval(timeout));
+  //  tqDebug("addTimeout %d", dbus_timeout_get_interval(timeout));
 
     TQT_DBusConnectionPrivate *d = static_cast<TQT_DBusConnectionPrivate *>(data);
 
@@ -77,7 +77,7 @@ static void qDBusRemoveTimeout(DBusTimeout *timeout, void *data)
     Q_ASSERT(timeout);
     Q_ASSERT(data);
 
-  //  qDebug("removeTimeout");
+  //  tqDebug("removeTimeout");
 
     TQT_DBusConnectionPrivate *d = static_cast<TQT_DBusConnectionPrivate *>(data);
     for (TQValueList<DBusTimeout*>::iterator it = d->pendingTimeouts.begin();
@@ -107,7 +107,7 @@ static void qDBusToggleTimeout(DBusTimeout *timeout, void *data)
     Q_ASSERT(timeout);
     Q_ASSERT(data);
 
-    //qDebug("ToggleTimeout");
+    //tqDebug("ToggleTimeout");
 
     qDBusRemoveTimeout(timeout, data);
     qDBusAddTimeout(timeout, data);
@@ -126,7 +126,7 @@ static dbus_bool_t qDBusAddWatch(DBusWatch *watch, void *data)
     TQT_DBusConnectionPrivate::Watcher watcher;
     if (flags & DBUS_WATCH_READABLE) {
         bool enabled = dbus_watch_get_enabled(watch);
-        //qDebug("addReadWatch %d %s", fd, (enabled ? "enabled" : "disabled"));
+        //tqDebug("addReadWatch %d %s", fd, (enabled ? "enabled" : "disabled"));
         watcher.watch = watch;
         if (tqApp) {
             watcher.read = new TQSocketNotifier(fd, TQSocketNotifier::Read, d);
@@ -136,7 +136,7 @@ static dbus_bool_t qDBusAddWatch(DBusWatch *watch, void *data)
     }
     if (flags & DBUS_WATCH_WRITABLE) {
         bool enabled = dbus_watch_get_enabled(watch);
-        //qDebug("addWriteWatch %d %s", fd, (enabled ? "enabled" : "disabled"));
+        //tqDebug("addWriteWatch %d %s", fd, (enabled ? "enabled" : "disabled"));
         watcher.watch = watch;
         if (tqApp) {
             watcher.write = new TQSocketNotifier(fd, TQSocketNotifier::Write, d);
@@ -160,7 +160,7 @@ static void qDBusRemoveWatch(DBusWatch *watch, void *data)
     Q_ASSERT(watch);
     Q_ASSERT(data);
 
-    //qDebug("remove watch");
+    //tqDebug("remove watch");
 
     TQT_DBusConnectionPrivate *d = static_cast<TQT_DBusConnectionPrivate *>(data);
     int fd = dbus_watch_get_unix_fd(watch);
@@ -202,7 +202,7 @@ static void qDBusToggleWatch(DBusWatch *watch, void *data)
     Q_ASSERT(watch);
     Q_ASSERT(data);
 
-    //qDebug("toggle watch");
+    //tqDebug("toggle watch");
 
     TQT_DBusConnectionPrivate *d = static_cast<TQT_DBusConnectionPrivate *>(data);
     int fd = dbus_watch_get_unix_fd(watch);
@@ -217,7 +217,7 @@ static void qDBusToggleWatch(DBusWatch *watch, void *data)
                 bool enabled = dbus_watch_get_enabled(watch);
                 int flags = dbus_watch_get_flags(watch);
 
-//                 qDebug("toggle watch %d to %d (write: %d, read: %d)",
+//                 tqDebug("toggle watch %d to %d (write: %d, read: %d)",
 //                         dbus_watch_get_unix_fd(watch), enabled,
 //                         flags & DBUS_WATCH_WRITABLE, flags & DBUS_WATCH_READABLE);
 
@@ -235,7 +235,7 @@ static void qDBusNewConnection(DBusServer *server, DBusConnection *c, void *data
 {
     Q_ASSERT(data); Q_ASSERT(server); Q_ASSERT(c);
 
-    qDebug("SERVER: GOT A NEW CONNECTION"); // TODO
+    tqDebug("SERVER: GOT A NEW CONNECTION"); // TODO
 }
 
 static DBusHandlerResult qDBusSignalFilter(DBusConnection *connection,
@@ -252,7 +252,7 @@ static DBusHandlerResult qDBusSignalFilter(DBusConnection *connection,
     bool handled = false;
 
     //TQT_DBusMessage amsg = TQT_DBusMessage::fromDBusMessage(message);
-    //qDebug() << "got message: " << dbus_message_get_type(message) << amsg;
+    //tqDebug() << "got message: " << dbus_message_get_type(message) << amsg;
 
     if (msgType == DBUS_MESSAGE_TYPE_SIGNAL) {
         handled = d->handleSignal(message);
@@ -273,7 +273,7 @@ int TQT_DBusConnectionPrivate::registerMessageMetaType()
 
 TQT_DBusConnectionPrivate::TQT_DBusConnectionPrivate(TQObject *parent)
     : TQObject(parent), ref(1), mode(InvalidMode), connection(0), server(0),
-      dispatcher(0)
+      dispatcher(0), inDispatch(false)
 {
     static const int msgType = registerMessageMetaType();
     Q_UNUSED(msgType);
@@ -282,6 +282,11 @@ TQT_DBusConnectionPrivate::TQT_DBusConnectionPrivate(TQObject *parent)
 
     dispatcher = new TQTimer(this);
     TQObject::connect(dispatcher, TQT_SIGNAL(timeout()), this, TQT_SLOT(dispatch()));
+
+    m_resultEmissionQueueTimer = new TQTimer(this);
+    TQObject::connect(m_resultEmissionQueueTimer, TQT_SIGNAL(timeout()), this, TQT_SLOT(transmitResultEmissionQueue()));
+    m_messageEmissionQueueTimer = new TQTimer(this);
+    TQObject::connect(m_messageEmissionQueueTimer, TQT_SIGNAL(timeout()), this, TQT_SLOT(transmitMessageEmissionQueue()));
 }
 
 TQT_DBusConnectionPrivate::~TQT_DBusConnectionPrivate()
@@ -363,7 +368,7 @@ void TQT_DBusConnectionPrivate::socketRead(int fd)
         for (WatcherList::const_iterator wit = list.begin(); wit != list.end(); ++wit) {
             if ((*wit).read && (*wit).read->isEnabled()) {
                 if (!dbus_watch_handle((*wit).watch, DBUS_WATCH_READABLE))
-                    qDebug("OUT OF MEM");
+                    tqDebug("OUT OF MEM");
             }
         }
     }
@@ -380,7 +385,7 @@ void TQT_DBusConnectionPrivate::socketWrite(int fd)
         for (WatcherList::const_iterator wit = list.begin(); wit != list.end(); ++wit) {
             if ((*wit).write && (*wit).write->isEnabled()) {
                 if (!dbus_watch_handle((*wit).watch, DBUS_WATCH_WRITABLE))
-                    qDebug("OUT OF MEM");
+                    tqDebug("OUT OF MEM");
             }
         }
     }
@@ -388,7 +393,7 @@ void TQT_DBusConnectionPrivate::socketWrite(int fd)
 
 void TQT_DBusConnectionPrivate::objectDestroyed(TQObject* object)
 {
-    //qDebug("Object destroyed");
+    //tqDebug("Object destroyed");
     for (PendingCallMap::iterator it = pendingCalls.begin(); it != pendingCalls.end();)
     {
         TQObject* receiver = (TQObject*) it.data()->receiver;
@@ -452,6 +457,13 @@ void TQT_DBusConnectionPrivate::scheduleDispatch()
 
 void TQT_DBusConnectionPrivate::dispatch()
 {
+    // dbus_connection_dispatch will hang if called recursively
+    if (inDispatch) {
+        printf("[dbus-1-tqt] WARNING: Attempt to call dispatch() recursively was silently ignored to prevent lockup!\n\r"); fflush(stdout);
+        return;
+    }
+    inDispatch = true;
+
     if (mode == ClientMode)
     {
         if (dbus_connection_dispatch(connection) != DBUS_DISPATCH_DATA_REMAINS)
@@ -460,6 +472,19 @@ void TQT_DBusConnectionPrivate::dispatch()
             dispatcher->stop();
         }
     }
+
+    inDispatch = false;
+}
+
+void TQT_DBusConnectionPrivate::transmitMessageEmissionQueue()
+{
+    TQT_DBusConnectionPrivate::PendingMessagesForEmit::iterator pmfe;
+    pmfe = pendingMessages.begin();
+    while (pmfe != pendingMessages.end()) {
+        TQT_DBusMessage msg = *pmfe;
+        pmfe = pendingMessages.remove(pmfe);
+        dbusSignal(msg);
+    }
 }
 
 bool TQT_DBusConnectionPrivate::handleObjectCall(DBusMessage *message)
@@ -481,7 +506,12 @@ bool TQT_DBusConnectionPrivate::handleSignal(DBusMessage *message)
     // FIXME-QT4
     //return handleSignal(TQString(), msg) | handleSignal(msg.path(), msg);
 
-    dbusSignal(msg);
+    // If dbusSignal(msg) were called here, it could easily cause a lockup as it would enter the TQt3 event loop,
+    // which could result in arbitrary methods being called while still inside dbus_connection_dispatch.
+    // Instead, I enqueue the messages here for TQt3 event loop transmission after dbus_connection_dispatch is finished.
+    pendingMessages.append(msg);
+    if (!m_messageEmissionQueueTimer->isActive()) m_messageEmissionQueueTimer->start(0, TRUE);
+
     return true;
 }
 
@@ -547,17 +577,17 @@ void TQT_DBusConnectionPrivate::setConnection(DBusConnection *dbc)
             return;
         }
     } else {
-        qWarning("TQT_DBusConnectionPrivate::SetConnection: Unable to get unique name");
+        tqWarning("TQT_DBusConnectionPrivate::SetConnection: Unable to get unique name");
     }
 
     dbus_connection_add_filter(connection, qDBusSignalFilter, this, 0);
 
-    //qDebug("unique name: %s", service);
+    //tqDebug("unique name: %s", service);
 }
 
 static void qDBusResultReceived(DBusPendingCall *pending, void *user_data)
 {
-    //qDebug("Pending Call Result received");
+    //tqDebug("Pending Call Result received");
     TQT_DBusConnectionPrivate* d = reinterpret_cast<TQT_DBusConnectionPrivate*>(user_data);
     TQT_DBusConnectionPrivate::PendingCallMap::iterator it = d->pendingCalls.find(pending);
 
@@ -570,13 +600,12 @@ static void qDBusResultReceived(DBusPendingCall *pending, void *user_data)
     {
         TQT_DBusMessage reply = TQT_DBusMessage::fromDBusMessage(dbusReply);
 
-        TQObject::connect(d, TQT_SIGNAL(dbusPendingCallReply(const TQT_DBusMessage&)),
-                         it.data()->receiver, it.data()->method.data());
-
-        d->emitPendingCallReply(reply);
-
-        TQObject::disconnect(d, TQT_SIGNAL(dbusPendingCallReply(const TQT_DBusMessage&)),
-                            it.data()->receiver, it.data()->method.data());
+        TQT_DBusResultInfo dbusResult;
+        dbusResult.message = reply;
+        dbusResult.receiver = it.data()->receiver;
+        dbusResult.method = it.data()->method.data();
+        d->m_resultEmissionQueue.append(dbusResult);
+        d->newMethodInResultEmissionQueue();
     }
 
     dbus_message_unref(dbusReply);
@@ -623,4 +652,26 @@ void TQT_DBusConnectionPrivate::flush()
     dbus_connection_flush(connection);
 }
 
+void TQT_DBusConnectionPrivate::newMethodInResultEmissionQueue()
+{
+    if (!m_resultEmissionQueueTimer->isActive()) m_resultEmissionQueueTimer->start(0, TRUE);
+}
+
+void TQT_DBusConnectionPrivate::transmitResultEmissionQueue()
+{
+    if (!m_resultEmissionQueue.isEmpty()) {
+        TQT_DBusResultInfoList::Iterator it;
+        it = m_resultEmissionQueue.begin();
+        while (it != m_resultEmissionQueue.end()) {
+            TQT_DBusResultInfo dbusResult = (*it);
+            m_resultEmissionQueue.remove(it);
+            it = m_resultEmissionQueue.begin();
+
+            TQObject::connect(this, TQT_SIGNAL(dbusPendingCallReply(const TQT_DBusMessage&)), dbusResult.receiver, dbusResult.method.data());
+            emitPendingCallReply(dbusResult.message);
+            TQObject::disconnect(this, TQT_SIGNAL(dbusPendingCallReply(const TQT_DBusMessage&)), dbusResult.receiver, dbusResult.method.data());
+        }
+    }
+}
+
 #include "tqdbusconnection_p.moc"
